@@ -18,6 +18,9 @@
 
 #define _GNU_SOURCE
 #include <string.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+
 #include <lttng/ust-config.h>
 #include <lttng/ust-ctl.h>
 #include <lttng/ust-abi.h>
@@ -109,8 +112,10 @@ int ustctl_release_object(int sock, struct lttng_ust_object_data *data)
 				ret = -errno;
 				return ret;
 			}
+			data->u.channel.wakeup_fd = -1;
 		}
 		free(data->u.channel.data);
+		data->u.channel.data = NULL;
 		break;
 	case LTTNG_UST_OBJECT_TYPE_STREAM:
 		if (data->u.stream.shm_fd >= 0) {
@@ -119,6 +124,7 @@ int ustctl_release_object(int sock, struct lttng_ust_object_data *data)
 				ret = -errno;
 				return ret;
 			}
+			data->u.stream.shm_fd = -1;
 		}
 		if (data->u.stream.wakeup_fd >= 0) {
 			ret = close(data->u.stream.wakeup_fd);
@@ -126,6 +132,7 @@ int ustctl_release_object(int sock, struct lttng_ust_object_data *data)
 				ret = -errno;
 				return ret;
 			}
+			data->u.stream.wakeup_fd = -1;
 		}
 		break;
 	case LTTNG_UST_OBJECT_TYPE_EVENT:
@@ -1791,6 +1798,105 @@ int ustctl_has_perf_counters(void)
 
 #endif
 
+#ifdef __linux__
+/*
+ * Override application pid/uid/gid with unix socket credentials. If
+ * the application announced a pid matching our view, it means it is
+ * within the same pid namespace, so expose the ppid provided by the
+ * application.
+ */
+static
+int get_cred(int sock,
+	const struct ustctl_reg_msg *reg_msg,
+	uint32_t *pid,
+	uint32_t *ppid,
+	uint32_t *uid,
+	uint32_t *gid)
+{
+	struct ucred ucred;
+	socklen_t ucred_len = sizeof(struct ucred);
+	int ret;
+
+	ret = getsockopt(sock, SOL_SOCKET, SO_PEERCRED, &ucred, &ucred_len);
+	if (ret) {
+		return -LTTNG_UST_ERR_PEERCRED;
+	}
+	DBG("Unix socket peercred [ pid: %u, uid: %u, gid: %u ], "
+		"application registered claiming [ pid: %u, ppid: %u, uid: %u, gid: %u ]",
+		ucred.pid, ucred.uid, ucred.gid,
+		reg_msg->pid, reg_msg->ppid, reg_msg->uid, reg_msg->gid);
+	if (!ucred.pid) {
+		ERR("Unix socket credential pid=0. Refusing application in distinct, non-nested pid namespace.");
+		return -LTTNG_UST_ERR_PEERCRED_PID;
+	}
+	*pid = ucred.pid;
+	*uid = ucred.uid;
+	*gid = ucred.gid;
+	if (ucred.pid == reg_msg->pid) {
+		*ppid = reg_msg->ppid;
+	} else {
+		*ppid = 0;
+	}
+	return 0;
+}
+#elif defined(__FreeBSD__)
+#include <sys/ucred.h>
+#include <sys/un.h>
+
+/*
+ * Override application uid/gid with unix socket credentials. Use the
+ * first group of the cr_groups.
+ * Use the pid and ppid provided by the application on registration.
+ */
+static
+int get_cred(int sock,
+	const struct ustctl_reg_msg *reg_msg,
+	uint32_t *pid,
+	uint32_t *ppid,
+	uint32_t *uid,
+	uint32_t *gid)
+{
+	struct xucred xucred;
+	socklen_t xucred_len = sizeof(struct xucred);
+	int ret;
+
+	ret = getsockopt(sock, SOL_SOCKET, LOCAL_PEERCRED, &xucred, &xucred_len);
+	if (ret) {
+		return -LTTNG_UST_ERR_PEERCRED;
+	}
+	if (xucred.cr_version != XUCRED_VERSION || xucred.cr_ngroups < 1) {
+		return -LTTNG_UST_ERR_PEERCRED;
+	}
+	DBG("Unix socket peercred [ uid: %u, gid: %u ], "
+		"application registered claiming [ pid: %d, ppid: %d, uid: %u, gid: %u ]",
+		xucred.cr_uid, xucred.cr_groups[0],
+		reg_msg->pid, reg_msg->ppid, reg_msg->uid, reg_msg->gid);
+	*pid = reg_msg->pid;
+	*ppid = reg_msg->ppid;
+	*uid = xucred.cr_uid;
+	*gid = xucred.cr_groups[0];
+	return 0;
+}
+#else
+#warning "Using insecure fallback: trusting user id provided by registered applications. Please consider implementing use of unix socket credentials on your platform."
+static
+int get_cred(int sock,
+	const struct ustctl_reg_msg *reg_msg,
+	uint32_t *pid,
+	uint32_t *ppid,
+	uint32_t *uid,
+	uint32_t *gid)
+{
+	DBG("Application registered claiming [ pid: %u, ppid: %d, uid: %u, gid: %u ]",
+		reg_msg->pid, reg_msg->ppid, reg_msg->uid, reg_msg->gid);
+	*pid = reg_msg->pid;
+	*ppid = reg_msg->ppid;
+	*uid = reg_msg->uid;
+	*gid = reg_msg->gid;
+	return 0;
+}
+#endif
+
 /*
  * Returns 0 on success, negative error value on error.
  */
@@ -1841,10 +1947,6 @@ int ustctl_recv_reg_msg(int sock,
 	}
 	*major = reg_msg.major;
 	*minor = reg_msg.minor;
-	*pid = reg_msg.pid;
-	*ppid = reg_msg.ppid;
-	*uid = reg_msg.uid;
-	*gid = reg_msg.gid;
 	*bits_per_long = reg_msg.bits_per_long;
 	*uint8_t_alignment = reg_msg.uint8_t_alignment;
 	*uint16_t_alignment = reg_msg.uint16_t_alignment;
@@ -1855,8 +1957,7 @@ int ustctl_recv_reg_msg(int sock,
 	if (reg_msg.major != LTTNG_UST_ABI_MAJOR_VERSION) {
 		return -LTTNG_UST_ERR_UNSUP_MAJOR;
 	}
-
-	return 0;
+	return get_cred(sock, &reg_msg, pid, ppid, uid, gid);
 }
 
 int ustctl_recv_notify(int sock, enum ustctl_notify_cmd *notify_cmd)
