@@ -943,10 +943,23 @@ int handle_message(struct sock_info *sock_info,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
+		if (args.channel.wakeup_fd >= 0) {
+			int close_ret;
+
+			lttng_ust_lock_fd_tracker();
+			close_ret = close(args.channel.wakeup_fd);
+			lttng_ust_unlock_fd_tracker();
+			args.channel.wakeup_fd = -1;
+			if (close_ret)
+				PERROR("close");
+		}
+		free(args.channel.chan_data);
 		break;
 	}
 	case LTTNG_UST_STREAM:
 	{
+		int close_ret;
+
 		/* Receive shm_fd, wakeup_fd */
 		ret = ustcomm_recv_stream_from_sessiond(sock,
 			NULL,
@@ -962,6 +975,22 @@ int handle_message(struct sock_info *sock_info,
 					&args, sock_info);
 		else
 			ret = -ENOSYS;
+		if (args.stream.shm_fd >= 0) {
+			lttng_ust_lock_fd_tracker();
+			close_ret = close(args.stream.shm_fd);
+			lttng_ust_unlock_fd_tracker();
+			args.stream.shm_fd = -1;
+			if (close_ret)
+				PERROR("close");
+		}
+		if (args.stream.wakeup_fd >= 0) {
+			lttng_ust_lock_fd_tracker();
+			close_ret = close(args.stream.wakeup_fd);
+			lttng_ust_unlock_fd_tracker();
+			args.stream.wakeup_fd = -1;
+			if (close_ret)
+				PERROR("close");
+		}
 		break;
 	}
 	case LTTNG_UST_CONTEXT:
